/* **********
 *
 * Boss and Worker classes
 *
 * Copyright (c) 2003 Dan B Goldman, Aaron Hertzmann
 * 
 * This file is part of the BossWorker library.
 *
 * BossWorker is free software; you can redistribute it and/or modify
 * it under the terms of the X11 Software License (see the LICENSE file
 * for details).
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANY; without even the implied warranty of
 * MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the X11
 * License for more details.
 *
 * **********/

#ifndef _BOSS_WORKER_H_
#define _BOSS_WORKER_H_

/* **********

   I'm calling this model a "boss/worker" API instead of
   client/server, because technically there is only one "client" and
   multiple "servers", which makes the terminology confusing.

   The boss process should be started first.  When a boss is created,
   he starts a hiring thread that waits for connections from any
   number of workers.  The connections are collected in a queue.

   As jobs (strings, actually) become available, they are sent to the
   workers immediately.  A separate thread for each worker waits for
   the responses to return.  As they do, the responses are placed in a
   mutex queue for collection by the initial thread.  From the
   perspective of the boss' client, he simply gives the boss jobs and
   receives the responses.  If no responses are available, the receive
   function blocks until a response appears.  Since the "receive" call
   is blocking, it's most efficient to send all the jobs in a single
   loop, then receive all the responses in another loop.

   Each job sent to a worker is stored by the Boss until the job is
   completed.  If a worker dies before completing its job, the job
   will be resent to another worker.

   Worker processes are single-threaded.  When they start up they
   initiate a connection to the boss.  They loop through receiving and
   processing jobs indefinitely, or until they are "fired" by the
   boss.

   NOTE: If no workers are available, then Boss::send function will
   block indefinitely.  If no responses are available, then
   Boss::receive will block indefinitely.  To avoid deadlock, it is
   important that at least one worker is available, that Workers must
   either send responses or die, and that the number of Boss::receive
   calls is identical to the number of Boss::send calls.
   
********** */

#include <string>
#include <vector>
#include <pthread.h>
#include "SyncQueue.h"
#include "Message.h"

typedef SyncQueue<std::string> ResponseQue;

class BossThread;

typedef std::vector<BossThread*> ThreadVec;
typedef SyncQueue<BossThread*> ThreadQue;

class Boss {
  private:
    int port;
    int listeningSocket;
    pthread_attr_t threadAttr;
    ThreadVec threads;
    ThreadQue connections;
    ResponseQue responses;

    int totalWorkers;
    int totalMessagesSent;
    int totalMessagesReceived;
    int totalBytesSent;
    int totalBytesReceived;

    static void *hireWorkers(void *p);  // entry point for hiring thread
    void hireWorkers();                 // member function for hiring thread

    friend class BossThread;

  public:
    Boss(int port);
    ~Boss();

    bool send(const std::string &msg);  // send a job to next available worker
    bool receive(std::string &msg);     // receive next available response

    /* convenience functions for using message classes instead of strings */

    inline bool send(const JobMessage &msg)
      {return send(msg.pack());}
    inline bool receive(ResultMessage &msg)
      {std::string s; bool b = receive(s); msg.unpack(s); return b;}

    std::string status();
        
};

class Worker {
  private:
    int socket;

    int totalMessagesSent;
    int totalMessagesReceived;
    int totalBytesSent;
    int totalBytesReceived;

  public:
    Worker(const std::string &bossName, int port);
    ~Worker();

    bool receive(std::string &msg);     // receive next available job
    bool send(const std::string &msg);  // return response immediately

    /* convenience functions for using message classes instead of strings */

    inline bool send(const ResultMessage &msg)
      {return send(msg.pack());}
    inline bool receive(JobMessage &msg)
      {std::string s; bool b = receive(s); msg.unpack(s); return b;}

    std::string status();

};

#endif // _BOSS_WORKER_
